# flake8: noqa
try:
    import bson
except ImportError:  # pragma: no cover
    bson = False

try:
    import dateutil
    import dateutil.parser
except ImportError:  # pragma: no cover
    dateutil = False

try:
    import marshmallow
except ImportError:  # pragma: no cover
    marshmallow = False

try:
    import pymongo
except ImportError:  # pragma: no cover
    pymongo = False

from flask_stupe.app import Stupeflask
from flask_stupe.validation import *
from flask_stupe.pagination import *
from flask_stupe.auth import *
from flask_stupe.inputs import *



import os
from pkgutil import iter_modules

from flask import Blueprint, Flask

from flask_stupe.config import Config
from flask_stupe.converters import converters
from flask_stupe.logging import log
from flask_stupe.request import Request


class Stupeflask(Flask):
    config_class = Config
    request_class = Request

    def __init__(self, *args, **kwargs):
        super(Stupeflask, self).__init__(*args, **kwargs)
        config_path = os.path.join(os.getcwd(), "config.py")
        log.info(" * Loading default config ({})".format(config_path))
        self.config.from_pyfile(config_path, silent=True)
        log.info(" * Loading $CONFIG ({})".format(os.environ.get("CONFIG")))
        self.config.from_envvar("CONFIG", silent=True)
        from_env = self.config.from_env()
        log.info(" * Overriden by environment: " + ", ".join(from_env))
        self.register_converters(converters)

    def register_converter(self, converter, name=None):
        """Register a new converter that can be used in endpoints URLs

        A converter is a Werkzeug class that transforms a part of an URL to a
        Python object, passed as an argument to the endpoint function.

        Stupeflask currently implements the following converters:

        - /<ObjectId>/ with ::ObjectIdConverter

        See `Werkzeug documentation about custom converters
        <http://werkzeug.pocoo.org/docs/0.10/routing/#custom-converters>`_
        if you want to implement your own converter.
        """
        if not name:
            name = converter.__name__
            if "Converter" in name:
                name = converter.__name__.replace("Converter", "")
        self.url_map.converters[name] = converter

    def register_converters(self, converter_list):
        """Register multiple converters at once.

        See :meth:`register_converter`.
        """
        for converter in converter_list:
            self.register_converter(converter)

    def register_blueprints(self, package):
        """Auto-discover blueprints and register them

        It will look recursively in every module of the package to find a
        variable that has the same name as its module.

        It means that if a file called `foo.py` has an exportable variable
        called `foo`, it will try to register that variable as a blueprint.
        """
        prefix = package.__name__ + '.'
        for importer, name, is_pkg in iter_modules(package.__path__, prefix):
            module = importer.find_module(name).load_module(name)
            blueprint_name = name.rsplit('.')[-1]
            blueprint = getattr(module, blueprint_name, None)
            if blueprint and isinstance(blueprint, Blueprint):
                log.info(' * Registering blueprint {}'.format(name))
                self.register_blueprint(blueprint)
            elif is_pkg:
                self.register_blueprints(module)


__all__ = ["Stupeflask"]



from datetime import date, datetime

from flask import jsonify, request
from flask import Response as FlaskResponse
from flask.json import JSONEncoder as FlaskJSONEncoder
from werkzeug.exceptions import (default_exceptions, HTTPException,
                                 InternalServerError)

from flask_stupe.app import Stupeflask as BaseStupeflask
from flask_stupe import bson

encoder_rules = [
    ((date, datetime), lambda d: d.isoformat()),
]


if bson:
    encoder_rules.append((bson.ObjectId, lambda o: str(o)))


class EncodeError(Exception):

    def __init__(self, o):
        self.message = "{} is not JSON serializable".format(repr(o))


def encode(o, silent=True):
    """Recursively encode a Python object in a JSON serializable format.

    objects natively supported by json's library are obviously supported, but
    so are datetimes, and ObjectId (if bson is installed).

    If you want to add such encoding for custom types, see
    :meth:`JSONEncoder.add_rule`.
    """
    for rule in encoder_rules:
        if isinstance(o, rule[0]):
            return rule[1](o)
    if not silent:
        raise EncodeError(o)
    return o


class JSONEncoder(FlaskJSONEncoder):
    """Stupeflask extensible JSON encoder

    The JSONEncoder is used by :func:`flask.jsonify` to convert a Python
    dictionary into a real JSON string. If the JSONEncoder receive a dict
    containing unknown type, a `TypeError` exception will be raised.

    Default supported types are Python built-in types (`int`, `list`, etc.),
    `datetime.date` and `datetime.date`, as well as `bson.ObjectId` if `bson`
    module is present.

    To make the JSONEncoder support more types, see :meth:`add_rule`.
    """

    def default(self, o):
        try:
            return encode(o, silent=False)
        except EncodeError:
            return super(JSONEncoder, self).default(o)

    def add_rule(type_or_tuple, function):
        """Add a new serializing rule

        A rule is defined by one or multiple types, and a function to execute
        in order to serialize this or those types.

        Dumb example:
            >>> app = flask.ext.stupe.json.Stupeflask(__name__)
            >>> app.json_encoder.add_rule(int, lambda i: str(i))
        """
        encoder_rules.append((type_or_tuple, function))


class Response(FlaskResponse):
    default_mimetype = "application/json"


def handle_error(e):
    """Convert any exception into a JSON message, with a proper HTTP code

    The JSON message will have the following form:

    .. code-block:: json

        {
            "code": 403,
            "message": "You don't have the permission to access the requested
                        resource. It is either read-protected or not readable
                        by the server."
        }
    """
    if not isinstance(e, HTTPException):
        e = InternalServerError()
    response = jsonify(code=e.code, message=e.description)
    response.status_code = e.code
    return response


class Stupeflask(BaseStupeflask):
    json_encoder = JSONEncoder
    response_class = Response

    def make_response(self, rv):
        if isinstance(rv, self.response_class):
            return rv

        data = None
        code = 200
        if isinstance(rv, tuple):
            data, code = rv
        elif isinstance(rv, int):
            code = rv
        else:
            data = rv

        rv = {"code": code}
        if data is not None:
            rv.update(data=data)
        if request.metadata:
            rv.update(**request.metadata)

        rv = jsonify(rv)
        rv.status_code = code
        return rv

    def __init__(self, *args, **kwargs):
        super(Stupeflask, self).__init__(*args, **kwargs)

        for code in default_exceptions.keys():
            self.register_error_handler(code, handle_error)


__all__ = ["encoder_rules", "encode", "JSONEncoder", "handle_error",
           "Stupeflask"]



import functools
import fnmatch

from flask import abort, request


def auth_required(function):
    """Decorator checking that the request is made by an authenticated user.

    If you want to use that function, you should set a before_request handler
    that authenticate requests when possible. It must then expose a `user`
    attribute on the :obj:`flask.request` object.

    .. code-block:: python

        @app.before_request
        def get_user():
            token = request.args.get("token")
            if verify_token(token):
                request.user = {"username": "toto"}

    A view decorated with :func:`auth_required` will be aborted with a status
    code 401 if the user making the request is not authenticated.
    """
    @functools.wraps(function)
    def __inner(*args, **kwargs):
        if not request.user:
            abort(401)
        return function(*args, **kwargs)
    return __inner


def permission_required(*permissions):
    """Decorator checking that the request is made by a user with adequate
    privileges.

    This decorator also decorates the function with the :func:`auth_required`
    doecorator, thus the same constraints apply here.

    If the :obj:`flask.request.user` is a dictionnary, it must contains a
    "permissions" key. If it is an object, it must have "permissions"
    attribute.

    These permissions must be a list of strings. They will be checked against
    the strings passed as parameter to this decorator. At least one must match.

    You can apply this decorator multiple times if you need several permissions
    at once to access a view.
    """
    def __decorator(function):
        @functools.wraps(function)
        def __inner(*args, **kwargs):
            if isinstance(request.user, dict):
                user_permissions = request.user.get("permissions", [])
            else:
                user_permissions = getattr(request.user, "permissions", [])

            for user_group in user_permissions:
                if fnmatch.filter(permissions, user_group):
                    return function(*args, **kwargs)
            abort(403)
        return auth_required(__inner)
    return __decorator


__all__ = ["auth_required", "permission_required"]



import functools

from flask import request

from flask_stupe import pymongo

__all__ = []


if pymongo:
    def _paginate(cursor, skip=None, limit=None, sort=None, count=True):
        if count:
            request.metadata.update(count=cursor.count())

        skip = request.args.get("skip", skip, type=int)
        if skip is not None:
            cursor.skip(skip)
        limit = request.args.get("limit", limit, type=int)
        if limit is not None:
            cursor.limit(limit)

        sort = request.args.get("sort", sort)
        if sort:
            if not isinstance(sort, list):
                sort = sort.split(",")
            for index, item in enumerate(sort):
                if isinstance(item, str):
                    if item.startswith("-"):
                        item = (item[1:], pymongo.DESCENDING)
                        sort[index] = item
                if not isinstance(item, tuple):
                    sort[index] = (item, pymongo.ASCENDING)
            cursor.sort(sort)
        return cursor

    def paginate(function_or_cursor=None, skip=None, limit=None, sort=None,
                 count=True):
        """Apply pagination to the given MongoDB cursor or function"""
        if isinstance(function_or_cursor, pymongo.cursor.Cursor):
            return _paginate(function_or_cursor, skip, limit, sort, count)

        def __decorator(function):
            @functools.wraps(function)
            def __wrapper(*args, **kwargs):
                cursor = function(*args, **kwargs)
                return _paginate(cursor, skip, limit, sort, count)
            return __wrapper

        if function_or_cursor:
            return __decorator(function_or_cursor)
        return __decorator

    __all__.append("paginate")



from flask import Request as FlaskRequest


class Request(FlaskRequest):

    def __init__(self, *args, **kwargs):
        super(Request, self).__init__(*args, **kwargs)

        #: Store additionnal data about the request.
        self.metadata = {}


__all__ = ["Request"]



import functools

from flask import abort, request

from flask_stupe import marshmallow

__all__ = []


if marshmallow:
    def schema_required(schema):
        """Validate body of the request against the schema.

        Abort with a status code 400 if the schema yields errors."""
        def __inner(f):
            @functools.wraps(f)
            def __inner(*args, **kwargs):
                json = request.get_json(force=True)
                results = schema.load(json)
                if results.errors:
                    abort(400, results.errors)
                request.schema = results.data
                return f(*args, **kwargs)
            return __inner
        return __inner

    __all__.extend(["schema_required"])



import ast
import os

from flask.config import Config as FlaskConfig


# ramnes: we don't want DEBUG=0 to be casted as DEBUG=True
def _str2bool(v):
    return bool(ast.literal_eval(v))


def _str2list(v):
    return v.split(",")


class Config(FlaskConfig):

    def from_env(self):
        """Try to overload the config with the environment.

        Take each key in the config and, if it is present as an environment
        variable, replace the current value with the environment one, casting
        it in the appropriate type.
        """
        env_config = {}
        for key, value in self.items():
            if key in os.environ:
                cast = type(value)
                if cast is bool:
                    cast = _str2bool
                elif cast is list:
                    cast = _str2list
                env_config[key] = cast(os.environ.get(key))
        self.update(env_config)
        return env_config


__all__ = ["Config"]



# encoding: utf-8
from setuptools import setup


def get_description():
    with open("README.rst") as file:
        return file.read()


setup(
    name='Flask-Stupe',
    version='3.6.3',
    url='https://github.com/numberly/flask-stupe',
    license='MIT',
    author='Guillaume Gelin',
    author_email='ramnes@1000mercis.com',
    description='a.k.a. « Flask on steroids »',
    long_description=get_description(),
    packages=['flask_stupe'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=['Flask>=0.11'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from werkzeug.routing import BaseConverter

from flask_stupe import bson

converters = []


if bson:
    class ObjectIdConverter(BaseConverter):
        regex = r'[A-Fa-f0-9]{24}'

        def to_python(self, value):
            return bson.ObjectId(value)

        def to_url(self, value):
            return str(value)

    converters.append(ObjectIdConverter)

__all__ = [converter.__name__ for converter in converters]



from flask_stupe import bson, dateutil


def boolean(value):
    """Parse the string ``"true"`` or ``"false"`` as a boolean (case
    insensitive). Also accepts ``"1"`` and ``"0"`` as ``True``/``False``
    (respectively). If the input is from the request JSON body, the type is
    already a native python boolean, and will be passed through without
    further parsing.

    This function belongs to https://github.com/flask-restful/flask-restful
    It is licensed under the BSD 3-clause "New" or "Revised" License
    https://github.com/flask-restful/flask-restful/blob/master/LICENSE
    """
    if isinstance(value, bool):
        return value

    if not value:
        raise ValueError("boolean type must be non-null")
    value = value.lower()
    if value in ('true', '1',):
        return True
    if value in ('false', '0',):
        return False
    raise ValueError("Invalid literal for boolean(): {0}".format(value))


def str_list(value):
    """Convert a literal comma separated string to a string list."""
    if not value:
        return []
    return value.split(",")


def int_list(value):
    """Convert a numeric comma separated string to an int list."""
    return [int(str) for str in str_list(value)]


__all__ = ["boolean", "int_list", "str_list"]


if bson:
    def objectid_list(value):
        """Convert a hexadecimal comma separated string to an ObjectId list."""
        return [bson.ObjectId(element) for element in str_list(value)]

    __all__.append("objectid_list")


if dateutil:
    def datetime(value):
        """Convert a string to a datetime"""
        return dateutil.parser.parse(value)

    def date(value):
        """Convert as string to a date"""
        return datetime(value).date()

    __all__.extend(["datetime", "date"])



import ipaddress
import re

from flask_stupe import bson, marshmallow

__all__ = []


if marshmallow:
    hexcolor_re = re.compile(r"^#[0-9a-f]{6}$")

    class Color(marshmallow.fields.String):
        default_error_messages = {
            "invalid": "Not a valid color."
        }

        def _deserialize(self, value, attr, data):
            try:
                value = value.lower()
            except AttributeError:
                self.fail("type")
            if not hexcolor_re.match(value):
                self.fail("invalid")
            return value

    class Cron(marshmallow.fields.String):
        default_error_messages = {
            "invalid": "Not a valid cron expression.",
            "minute": "The minutes field is invalid.",
            "hour": "The hours field is invalid.",
            "dom": "The day of the month field is invalid.",
            "month": "The month field is invalid.",
            "dow": "The day of the week field is invalid."
        }
        limits = (("minute", 59), ("hour", 23), ("dom", 31), ("month", 12),
                  ("dow", 7))

        def _deserialize(self, value, attr, data):
            fields = value.split()
            if len(fields) != 5:
                self.fail("invalid")
            for field, (name, limit) in zip(fields, self.limits):
                try:
                    if field != "*":
                        if int(field) > limit or int(field) < 0:
                            raise Exception
                except ValueError:
                    self.fail("invalid")
                except Exception:
                    self.fail(name)
            return value

    currencies = ("ADF", "ADP", "AED", "AFA", "AFN", "ALL", "AMD", "ANG",
                  "AOA", "AOK", "AON", "AOR", "ARP", "ARS", "ATS", "AUD",
                  "AWG", "AZM", "AZN", "BAM", "BBD", "BDT", "BEF", "BGL",
                  "BGN", "BHD", "BIF", "BMD", "BND", "BOB", "BOP", "BOV",
                  "BRL", "BRR", "BSD", "BTN", "BWP", "BYB", "BYN", "BYR",
                  "BZD", "CAD", "CDF", "CHE", "CHF", "CHW", "CLF", "CLP",
                  "CNY", "COP", "COU", "CRC", "CSD", "CSK", "CUC", "CUP",
                  "CVE", "CYP", "CZK", "DEM", "DJF", "DKK", "DOP", "DZD",
                  "ECS", "ECV", "EEK", "EGP", "ERN", "ESP", "ETB", "EUR",
                  "FIM", "FJD", "FKP", "FRF", "GBP", "GEL", "GHS", "GIP",
                  "GMD", "GNF", "GRD", "GTQ", "GWP", "GYD", "HKD", "HNL",
                  "HRK", "HTG", "HUF", "IDR", "IEP", "ILS", "INR", "IQD",
                  "IRR", "ISK", "ITL", "JMD", "JOD", "JPY", "KES", "KGS",
                  "KHR", "KMF", "KPW", "KRW", "KWD", "KYD", "KZT", "LAK",
                  "LBP", "LKR", "LRD", "LSL", "LTL", "LUF", "LVL", "LVR",
                  "LYD", "MAD", "MDL", "MGA", "MGF", "MKD", "MMK", "MNT",
                  "MOP", "MRO", "MTL", "MUR", "MVR", "MWK", "MXN", "MXV",
                  "MYR", "MZE", "MZM", "MZN", "NAD", "NGN", "NHF", "NIC",
                  "NIO", "NLG", "NOK", "NPR", "NZD", "OMR", "PAB", "PEN",
                  "PES", "PGK", "PHP", "PKR", "PLN", "PLZ", "PTE", "PYG",
                  "QAR", "ROL", "RON", "RSD", "RUB", "RWF", "SAR", "SBD",
                  "SCR", "SDD", "SDG", "SDP", "SEK", "SGD", "SHP", "SIT",
                  "SKK", "SLL", "SML", "SOS", "SRD", "SSP", "STD", "SUB",
                  "SUR", "SVC", "SYP", "SZL", "THB", "TJS", "TMM", "TMT",
                  "TND", "TOP", "TPE", "TRL", "TRY", "TTD", "TWD", "TZS",
                  "UAH", "UGX", "USD", "USN", "USS", "UYU", "UZS", "VAL",
                  "VEB", "VEF", "VND", "VUV", "WST", "XAF", "XAG", "XAU",
                  "XBA", "XBB", "XBC", "XBD", "XCD", "XDR", "XEU", "XFO",
                  "XFU", "XOF", "XPD", "XPF", "XPT", "YER", "YUD", "YUM",
                  "ZAR", "ZMK", "ZWD", "ZWL", "ZWR")

    class Currency(marshmallow.fields.String):
        default_error_messages = {
            "invalid": "Not a valid currency."
        }

        def _deserialize(self, value, attr, data):
            try:
                value = value.upper()
            except AttributeError:
                self.fail("type")
            if value not in currencies:
                self.fail("invalid")
            return value

    class IP(marshmallow.fields.String):
        default_error_messages = {
            "invalid": "Not a valid IPv4 or IPv6 address."
        }
        ip_type = staticmethod(ipaddress.ip_address)

        def _deserialize(self, value, attr, data):
            try:
                self.ip_type(value)
            except ValueError:
                self.fail("invalid")
            return value

    class IPv4(IP):
        default_error_messages = {
            "invalid": "Not a valid IPv4 address."
        }
        ip_type = ipaddress.IPv4Address

    class IPv6(IP):
        default_error_messages = {
            "invalid": "Not a valid IPv6 address."
        }
        ip_type = ipaddress.IPv6Address

    class OneOf(marshmallow.fields.Field):
        """Validate a value against several fields until one match.

        :param fields: The list of fields to check.
        """
        default_error_messages = {
            "invalid": "Object type doesn't match any valid type"
        }

        def __init__(self, fields, *args, **kwargs):
            super(OneOf, self).__init__(*args, **kwargs)

            if not isinstance(fields, (list, tuple)):
                raise ValueError("Fields must be contained in a list or tuple")

            self.fields = []
            for field in fields:
                if isinstance(field, type):
                    if not issubclass(field, marshmallow.base.FieldABC):
                        raise ValueError("Fields types must subclass FieldABC")
                    self.fields.append(field())
                else:
                    if not isinstance(field, marshmallow.base.FieldABC):
                        raise ValueError("Fields must be FieldABC instances")
                    self.fields.append(field)

        def _deserialize(self, value, *args, **kwargs):
            for field in self.fields:
                try:
                    return field._deserialize(value, *args, **kwargs)
                except marshmallow.exceptions.ValidationError:
                    pass
            self.fail("invalid")

    __all__.extend(["Color", "Currency", "IP", "IPv4", "IPv6", "OneOf"])


if bson and marshmallow:
    class ObjectId(marshmallow.fields.String):
        default_error_messages = {
            "invalid": "Not a valid ObjectId."
        }

        def _deserialize(self, value, attr, data):
            try:
                return bson.ObjectId(value)
            except TypeError:
                self.fail("type")
            except bson.errors.InvalidId:
                self.fail("invalid")

    __all__.append("ObjectId")
